import { Logger, VersioningType } from '@nestjs/common';
import { NestFactory } from '@nestjs/core';
import { FastifyAdapter, NestFastifyApplication } from '@nestjs/platform-fastify';
import { Logger as PinoLogger } from 'nestjs-pino';

import { DocumentBuilder, SwaggerDocumentOptions, SwaggerModule } from '@nestjs/swagger';
import { flow } from 'lodash';
import { AppModule } from './app.module';
import { patchPublicDecoratorSupport } from './modules/auth/strategies/jwt/jwt-auth.guard';
import { LocalConfig, localConfig } from './modules/config/local.config';
import { patchTypeOrm } from './modules/database/utils/typeorm-patch';
import { patchNestjsSwagger } from './modules/validation/utils/nestjs-swagger-patches';

async function bootstrap() {
  await patchTypeOrm();
  await patchNestjsSwagger();

  const appOpts = (() => {
    const loggerOpts = { bufferLogs: true };
    return {
      forceCloseConnections: true,
      ...loggerOpts,
    };
  })();
  const app = await NestFactory.create<NestFastifyApplication>(AppModule, new FastifyAdapter(appOpts));
  app.enableVersioning({
    type: VersioningType.URI,
    defaultVersion: '1',
  });
  app.enableShutdownHooks();

  const config = app.get<LocalConfig>(localConfig.KEY);
  const port = config.port || 3000;

  app.useLogger(app.get(PinoLogger));

  const swaggerConfig = new DocumentBuilder().addBearerAuth().build();
  const swaggerOptions = {
    operationIdFactory: (controllerKey: string, methodKey: string) => {
      return `${controllerKey}_${methodKey}`;
    },
  } satisfies SwaggerDocumentOptions;
  const document = flow(
    () => SwaggerModule.createDocument(app, swaggerConfig, swaggerOptions),
    patchPublicDecoratorSupport,
  )();

  SwaggerModule.setup(`docs`, app, document, {
    swaggerOptions: {
      operationsSorter: 'alpha',
      displayOperationId: true,
      filter: true,
      persistAuthorization: true,
    },
  });

  await app.listen(port);
  Logger.log(`🚀 Application is running on: http://localhost:${port}`);
}

void bootstrap();
