import { Inject, Injectable, OnApplicationBootstrap } from '@nestjs/common';
import { differenceBy, get } from 'lodash';
import { DataSource, QueryFailedError } from 'typeorm';
import { APP_DATASOURCES } from '../database/database.constants';
import { Enumify, registeredEnums } from './enumify';
import { type EnumifyConfig, enumifyConfig } from './enumify.config';

@Injectable()
export class EnumifyUpdateFromDatabaseService implements OnApplicationBootstrap {
  constructor(
    @Inject(APP_DATASOURCES) private readonly datasources: DataSource[],
    @Inject(enumifyConfig.KEY) private readonly config: EnumifyConfig,
  ) {}

  async onApplicationBootstrap() {
    try {
      const enumUpdateFromDbOnStart = this.config.shouldUpdateEnumFromDbOnStartup;
      if (!enumUpdateFromDbOnStart) return;

      const errors: string[] = [];

      for (const dataSource of this.datasources) {
        const logging = dataSource.options.logging;
        dataSource.setOptions({ logging: false });
        for (const _registeredEnum of get(registeredEnums, dataSource.name, [])) {
          const registeredEnum = _registeredEnum;

          const repo = dataSource.getRepository(registeredEnum);

          const enumValues = registeredEnum.enumValues;
          let dbValues: Enumify[];
          try {
            dbValues = await repo.find();
          } catch (err) {
            if (
              err instanceof QueryFailedError &&
              err.message === `Error: Invalid object name '${repo.metadata.tableName}'.`
            ) {
              errors.push(
                `[${dataSource.name}] ${registeredEnum.name}: Table present in code but missing on database: ${enumValues}`,
              );
            }
            if (err instanceof QueryFailedError && err.message.startsWith('Error: Invalid column name')) {
              errors.push(`[${dataSource.name}] ${registeredEnum.name}: [${repo.metadata.tableName}] ${err.message}`);
            }
            continue;
          }

          const differenceByDbValues = differenceBy(dbValues, enumValues, (x) => (x.id, x.nome));
          const differenceByEnumValues = differenceBy(enumValues, dbValues, (x) => (x.id, x.nome));
          if (differenceByDbValues.length > 0) {
            errors.push(
              `[${dataSource.name}] ${
                registeredEnum.name
              }: Values present on database but missing in code: ${differenceBy(
                dbValues,
                enumValues,
                (x) => (x.id, x.nome),
              )}`,
            );
          }
          if (differenceByEnumValues.length > 0) {
            errors.push(
              `[${dataSource.name}] ${
                registeredEnum.name
              }: Values present in code but missing in database: ${differenceBy(
                enumValues,
                dbValues,
                (x) => (x.id, x.nome),
              )}`,
            );
          }

          for (const dbValue of dbValues) {
            const valueOfByName = registeredEnum.fromKey(dbValue.nome!);
            const keyOfByCode = registeredEnum.fromValue(dbValue.id!);
            if (valueOfByName != null && dbValue.id != valueOfByName?.id) {
              errors.push(
                `[${dataSource.name}] ${registeredEnum.name}: Different values between database (${dbValue.id}, ${dbValue.nome}) and code (${valueOfByName.id}, ${valueOfByName.nome})`,
              );
            } else if (keyOfByCode != null && dbValue.nome != keyOfByCode?.nome) {
              errors.push(
                `[${dataSource.name}] ${registeredEnum.name}: Different values between database (${dbValue.id}, ${dbValue.nome}) and code (${keyOfByCode.id}, ${keyOfByCode.nome})`,
              );
            } else if (valueOfByName != null || keyOfByCode != null) {
              const enumValue = (valueOfByName ?? keyOfByCode)!;
              Object.assign(enumValue, dbValue);
            }
          }
        }
        dataSource.setOptions({ logging: logging });
      }

      if (errors.length > 0) {
        throw new Error(errors.join('\n\t* '));
      }
    } catch (err) {
      if (err instanceof Error) {
        console.warn(err.message, EnumifyUpdateFromDatabaseService.name);
      } else {
        throw err;
      }
    }
  }
}
