import { Logger } from '@nestjs/common';
import fs from 'node:fs';
import { randomUUID } from 'node:crypto';
import path from 'node:path';
import { FileTransactionsService } from './file-transactions.service';
import { copyFilesToFolderOrFail, deleteFiles, renameFiles, renameFilesOrFail } from './file-utils';
import dayjs from 'dayjs';

export class FileTransaction {
  public static readonly TO_ADD_SUFFIX = 'toAdd';
  public static readonly TO_DELETE_SUFFIX = 'toDelete';

  private readonly manager: FileTransactionsService;
  private readonly logger: Logger;

  public readonly transactionId: string;
  private readonly ts: Date;
  public readonly filesToAdd: { src: string; dest: string }[];
  public readonly filesToDelete: string[];

  constructor(
    manager: FileTransactionsService,
    opts: Partial<{
      transactionId: string;
      filesToAdd: { src: string; dest: string }[];
      filesToDelete: string[];
      logger: Logger;
    }>,
  ) {
    this.manager = manager;
    this.logger = opts.logger ?? new Logger(FileTransaction.name);
    this.filesToAdd = opts.filesToAdd ?? [];
    this.filesToDelete = opts.filesToDelete ?? [];
    this.transactionId = opts.transactionId ?? (randomUUID().split('-').pop() as string);
    this.ts = new Date();
  }

  async addFile(src: string, dest: string) {
    const destFolderName = path.dirname(dest);
    const destFileName = path.basename(dest);
    const destFileNameWithSuffix = `${destFileName}.${this.transactionId}.${FileTransaction.TO_ADD_SUFFIX}`;
    const destFilePathWithSuffix = path.join(destFolderName, destFileNameWithSuffix);

    this.logger.log(
      `Adding file '${destFilePathWithSuffix}' in fileTransaction '${
        this.transactionId
      }' (started at ${dayjs(this.ts).format('YYYY-MM-DD HH:mm:ss')})`,
    );
    this.filesToAdd.push({ src: src, dest: destFilePathWithSuffix });

    if (fs.existsSync(dest)) {
      await this.deleteFile(dest);
    }

    await copyFilesToFolderOrFail(this.logger, destFolderName, {
      srcpath: src,
      name: destFileNameWithSuffix,
    });
    await deleteFiles(this.logger, src);
    this.manager.persist(this);
  }

  async deleteFile(src: string) {
    const destFolderName = path.dirname(src);
    const destFileName = path.basename(src);
    const destFileNameWithSuffix = `${destFileName}.${this.transactionId}.${FileTransaction.TO_DELETE_SUFFIX}`;
    const destFilePathWithSuffix = path.join(destFolderName, destFileNameWithSuffix);

    this.logger.log(
      `Deleting file '${destFilePathWithSuffix}' in fileTransaction '${
        this.transactionId
      }' (started at ${dayjs(this.ts).format('YYYY-MM-DD HH:mm:ss')})`,
    );
    this.filesToDelete.push(destFilePathWithSuffix);
    await renameFilesOrFail(this.logger, { src, dest: destFilePathWithSuffix });
    this.manager.persist(this);
  }

  async commitTransaction() {
    this.logger.log(
      `Committing fileTransaction '${this.transactionId}' (started at ${dayjs(this.ts).format('YYYY-MM-DD HH:mm:ss')})`,
    );
    await renameFiles(
      this.logger,
      ...this.filesToAdd.map((fileToAdd) => ({
        src: fileToAdd.dest,
        dest: fileToAdd.dest.replace(`.${this.transactionId}.${FileTransaction.TO_ADD_SUFFIX}`, ''),
      })),
    );
    await deleteFiles(this.logger, ...this.filesToDelete);
    this.manager.unregister(this);
  }

  async rollbackTransaction() {
    this.logger.log(
      `Rollbacking fileTransaction '${this.transactionId}' (started at ${dayjs(this.ts).format(
        'YYYY-MM-DD HH:mm:ss',
      )})`,
    );

    await renameFiles(
      this.logger,
      ...this.filesToAdd.map((fileToMove) => ({
        src: fileToMove.dest,
        dest: fileToMove.src,
      })),
    );

    await renameFiles(
      this.logger,
      ...this.filesToDelete.map((fileToRestore) => ({
        src: fileToRestore,
        dest: fileToRestore.replace(`.${this.transactionId}.${FileTransaction.TO_DELETE_SUFFIX}`, ''),
      })),
    );
    this.manager.unregister(this);
  }
}
