import { Injectable, Logger } from '@nestjs/common';
import { InjectRepository } from '@nestjs/typeorm';
import { isError } from '@stdlib/assert';
import AsyncLock from 'async-lock';
import { Keyv } from 'cacheable';
import { isObject } from 'lodash';
import { Repository } from 'typeorm';
import { runInTransaction } from 'typeorm-transactional';
import { ApiClientEntity, ApiClientInvocazioniEntity } from '../../database/connections/ebitemp-api';
import { EBITEMP_API_DATASOURCE } from '../../database/connections/ebitemp-api/database.constants';
import { KeyvService } from '../../keyv/keyv.service';

@Injectable()
export class ClientsService {
  private readonly logger = new Logger(ClientsService.name);
  private readonly lock = new AsyncLock();

  private readonly keyv: Keyv;

  constructor(
    keyvService: KeyvService,
    @InjectRepository(ApiClientEntity, EBITEMP_API_DATASOURCE)
    private readonly clientRepository: Repository<ApiClientEntity>,
    @InjectRepository(ApiClientInvocazioniEntity, EBITEMP_API_DATASOURCE)
    private readonly clientInvocazioniRepository: Repository<ApiClientInvocazioniEntity>,
  ) {
    this.keyv = keyvService.create({
      namespace: ClientsService.name,
    });
  }

  async insertOne(
    baseUrl: string,
    path: string,
    method: string,
    requestBody: any,
    responseBody: any,
    hasFailed: boolean,
    errors: string | null,
  ) {
    try {
      return await runInTransaction(
        async () => {
          const client = await this.getClient(baseUrl);
          if (!client) return;

          const payload = JSON.stringify(requestBody);
          const response = isObject(responseBody) ? JSON.stringify(responseBody) : responseBody;

          const invocazione = this.clientInvocazioniRepository.create({
            idClient: client.id,
            percorso: path,
            metodo: method,
            timestamp: new Date(),
            richiesta: payload,
            risposta: response,
            errori: errors,
            flagErrore: hasFailed,
          });

          await this.clientInvocazioniRepository.insert(invocazione);
          return invocazione;
        },
        { connectionName: EBITEMP_API_DATASOURCE },
      );
    } catch (err) {
      if (!isError(err)) throw err;
      this.logger.error(`Unexpected error: ${err.message}`);
      throw err;
    }
  }

  private async getClient(baseUrl: string) {
    const key = `client:baseUrl#${baseUrl}`;

    return await this.lock.acquire(key, async () => {
      let client = await this.keyv.get<ApiClientEntity>(key);
      const isInCache = client != null;
      if (!isInCache) {
        client ??=
          (await this.clientRepository.findOne({
            where: { baseUrl: baseUrl },
          })) ?? undefined;
        if (client) this.keyv.set(key, client);
      }
      return client;
    });
  }
}
