import { CallHandler, ExecutionContext, Injectable, Logger, NestInterceptor } from '@nestjs/common';
import { omit } from 'lodash';
import { Observable } from 'rxjs';
import { catchError, concatMap } from 'rxjs/operators';
import { EndpointsService } from './endpoints/endpoints.service';
import { RequestWithUser } from '../auth/constants/request-with-user';

export type RequestLog = {
  url: string;
  idAccount: string;
  params: object;
  query: object;
  body: object;
};

@Injectable()
export class RequestLoggerInterceptor implements NestInterceptor {
  private readonly logger = new Logger(RequestLoggerInterceptor.name);

  constructor(private readonly endpointsService: EndpointsService) {}

  async intercept(context: ExecutionContext, next: CallHandler): Promise<Observable<any>> {
    const request: RequestWithUser = context.switchToHttp().getRequest();

    const user = request.user;
    const route = request.routeOptions.url ?? '/';
    const method = request.method;

    request.params = { ...((request.params as object) ?? {}), route, method };

    const requestLog = {
      url: request.url,
      idAccount: `${user.id}`,
      params: omit(request.params as object, ['user', 'route', 'method']),
      query: omit(request.query as object, ['__context']),
      body: request.body as object,
    };

    return next.handle().pipe(
      concatMap(async (res: any) => {
        await this.endpointsService.insertOne(user, route, method, requestLog, res, false);
        return res;
      }),
      catchError(async (err: any) => {
        const getExceptionResponse = (exception: any): any => {
          if (typeof exception.response === 'object') {
            const { error, message, ...rest } = exception.response;
            return {
              error,
              message,
              ...rest,
            };
          }
          return {
            error: exception.name,
            message: exception.message,
          };
        };

        const res = getExceptionResponse(err);

        await this.endpointsService.insertOne(user, route, method, requestLog, res, true);
        throw err;
      }),
    );
  }
}
