import { ZodDtoStatic } from '@anatine/zod-nestjs';
import { HTTP_ERRORS_BY_CODE } from '@anatine/zod-nestjs/src/lib/http-errors';
import {
  CallHandler,
  ExecutionContext,
  HttpStatus,
  Inject,
  Injectable,
  NestInterceptor,
  Optional,
  SetMetadata,
  StreamableFile,
} from '@nestjs/common';
import { chain } from 'lodash';
import { Observable, map } from 'rxjs';
import { ZodSchema } from 'zod';

const REFLECTOR = 'Reflector';

export interface ZodSerializerInterceptorOptions {
  errorHttpStatusCode?: keyof typeof HTTP_ERRORS_BY_CODE;
}

export const ZodSerializerDto = (dto: ZodDtoStatic | ZodSchema) => SetMetadata('zodSerializedDtoOptions', { dto });

@Injectable()
export class ZodSerializerInterceptor implements NestInterceptor {
  private readonly errorHttpStatusCode: keyof typeof HTTP_ERRORS_BY_CODE;

  constructor(
    @Inject(REFLECTOR) protected readonly reflector: any,
    @Optional() options?: ZodSerializerInterceptorOptions,
  ) {
    this.errorHttpStatusCode = options?.errorHttpStatusCode || HttpStatus.INTERNAL_SERVER_ERROR;
  }

  intercept(context: ExecutionContext, next: CallHandler): Observable<any> {
    const responseSchema = this.getContextResponseSchema(context);

    return next.handle().pipe(
      map((res: object | object[]) => {
        if (!responseSchema) return res;
        if (typeof res !== 'object' || res instanceof StreamableFile) return res;
        if (!('safeParse' in responseSchema)) return res;

        return Array.isArray(res)
          ? res.map(
              (item) =>
                validateInterally(responseSchema, item, {
                  httpStatusCodeOnError: this.errorHttpStatusCode,
                }).data,
            )
          : validateInterally(responseSchema, res, {
              httpStatusCodeOnError: this.errorHttpStatusCode,
            }).data;
      }),
    );
  }

  protected getContextResponseSchema(context: ExecutionContext): ZodDtoStatic | ZodSchema | undefined {
    const zodSerializedDto = (() => {
      const metadata = this.reflector.getAllAndMerge('zodSerializedDtoOptions', [
        context.getHandler(),
        context.getClass(),
      ]);
      return metadata.dto;
    })();
    if (zodSerializedDto) return zodSerializedDto;

    const swaggerApiResponseType = (() => {
      const metadata = this.reflector.getAllAndMerge('swagger/apiResponse', [context.getHandler(), context.getClass()]);
      const metadataForStatusCode = chain([metadata[HttpStatus.OK], metadata[HttpStatus.CREATED]])
        .filter(Boolean)
        .first()
        .value();
      return metadataForStatusCode.type;
    })();
    if (swaggerApiResponseType) return swaggerApiResponseType;

    return undefined;
  }
}

function isZodDto(metatype: unknown): metatype is ZodDtoStatic<any> {
  return !!(metatype as ZodDtoStatic)?.zodSchema;
}

function validateInterally(
  schemaOrDto: ZodDtoStatic | ZodSchema,
  value: any,
  options: Partial<{
    httpStatusCodeOnError: keyof typeof HTTP_ERRORS_BY_CODE;
  }>,
) {
  const { httpStatusCodeOnError = HttpStatus.INTERNAL_SERVER_ERROR } = options;
  const schema = isZodDto(schemaOrDto) ? schemaOrDto.zodSchema : schemaOrDto;

  const parseResult = schema.safeParse(value);
  if (!parseResult.success) {
    const { error } = parseResult;
    const message = error.errors.map((error) => `${error.path.join('.')}: ${error.message}`);

    console.error(
      `Error while validating dto '${isZodDto(schemaOrDto) ? schemaOrDto.name : 'unknownSchema'}': ${message}`,
    );
    console.error('value', value);
    throw new HTTP_ERRORS_BY_CODE[httpStatusCodeOnError]();
  }
  return parseResult;
}
